
/*
  Copyright (C) 2002  Dmitry V. Levin <ldv@altlinux.org>

  appconf

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <errno.h>
#include <error.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/stat.h>
#include <pwd.h>

#include "shvar.h"
#include "server.h"

#define APPSDIR "/etc/security/console.apps/"

const char *__progname;

void
get_appconf(appconf * conf)
{
	char   *fname;
	shvarFile *s;
	const char *tmp;
	char   *program = NULL;
	struct stat stb;

	if (asprintf(&fname, APPSDIR "%s", __progname) < 0)
		error(EXIT_FAILURE, errno, "asprintf");

	s = svNewFile(fname);
	if (!s)
	{
		syslog(LOG_WARNING, "appconf: %s: %m", fname);
		exit(EXIT_FAILURE);
	}

	if ((fstat(s->fd, &stb) < 0) || !S_ISREG(stb.st_mode))
	{
		syslog(LOG_ERR, "appconf: %s: bad file type", fname);
		exit(EXIT_FAILURE);
	}

	if (stb.st_mode & (S_IWGRP | S_IWOTH))
	{
		syslog(LOG_ERR, "appconf: %s: bad file permissions", fname);
		exit(EXIT_FAILURE);
	}

	tmp = svGetValue(s, "USER");
	if (!tmp || !strcmp(tmp, "<user>"))
		conf->user = get_logname();
	 else
	 {
		conf->user = strdup(tmp);
		if (!conf->user)
			error(EXIT_FAILURE, errno, "strdup");
	 }

	if (!conf->user)
	{
		syslog(LOG_WARNING, "appconf: %s: undefined user", fname);
		exit(EXIT_FAILURE);
	}

	tmp = svGetValue(s, "RUNAS");
	{
		struct passwd *pw = tmp ? getpwnam(tmp) : getpwuid(0);

		if (!pw)
		{
			syslog(LOG_WARNING, "appconf: %s: unknown runas",
			       fname);
			exit(EXIT_FAILURE);
		}
		tmp = pw->pw_name;
	}
	conf->runas = strdup(tmp);
	if (!conf->runas)
		error(EXIT_FAILURE, errno, "strdup");

	tmp = svGetValue(s, "PROGRAM");
	if (!tmp || !tmp[0])
		tmp = __progname;

	program = strdup(tmp);
	if (!program)
		error(EXIT_FAILURE, errno, "strdup");

	if (program[0] == '/')
	{
		if (access(program, X_OK))
		{
			syslog(LOG_ERR,
			       "appconf: %s: missing program: %s",
			       fname, program);
			exit(EXIT_FAILURE);
		}
	} else
	{
		char   *path;

		if (asprintf(&path, "/sbin/%s", program) < 0)
			error(EXIT_FAILURE, errno, "asprintf");

		if (access(path, X_OK))
		{
			free(path);
			if (asprintf(&path, "/usr/sbin/%s", program) < 0)
				error(EXIT_FAILURE, errno, "asprintf");
			if (access(path, X_OK))
			{
				syslog(LOG_ERR,
				       "appconf: %s: missing program: %s",
				       fname, program);
				exit(EXIT_FAILURE);
			}
		}

		free(program);
		program = path;
	}

	if (!program)
	{
		syslog(LOG_WARNING, "appconf: %s: undefined program", fname);
		exit(EXIT_FAILURE);
	}
	conf->program = program;

	conf->session = svTrueValue(s, "SESSION", 0);

	svCloseFile(s);
}
