
/*
  Copyright (C) 2002, 2004  Dmitry V. Levin <ldv@altlinux.org>

  pam auth

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <errno.h>
#include <error.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <pwd.h>
#include <sys/wait.h>

#include "serv_pam.h"
#include <security/pam_misc.h>

#include "server.h"
#include "server.h"

#define	DEFAULT_HOME	"/"

extern const char *__progname;

static pam_handle_t *pamh;

static struct pam_conv conv = {
	do_pam_conversation,
	0
};

static void
__attribute__ ((__noreturn__))
__attribute__ ((__format__(__printf__, 1, 2)))
fail(const char *fmt, ...)
{
	va_list args;

	openlog("consolehelper", LOG_PID, LOG_AUTH);

	va_start(args, fmt);
	vsyslog(LOG_WARNING, fmt, args);
	va_end(args);

	if (pamh)
		pam_end(pamh, PAM_ABORT);
	pamh = 0;

	exit(EXIT_FAILURE);
}

static void
__attribute__ ((__noreturn__))
exec_program(uid_t uid, const appconf * conf)
{
	char   *const *shell_env;
	const char *pw_dir;
	int     rc;

	/*
	 * Now, find the home directory for the user
	 */

	pw_dir = pam_getenv(pamh, "HOME");
	if (!pw_dir || !*pw_dir)
	{
		/* Not set so far, so we get it now. */
		struct passwd *pwd;

		pwd = getpwnam(conf->user);
		if (pwd && pwd->pw_name)
			pw_dir = strdup(pwd->pw_name);

		if (!pw_dir || !*pw_dir)
			fail("error setting home directory for %s",
			     conf->user);
	}

	/*
	 * We wish to change the current directory to smth sane.
	 */

	if (chdir(DEFAULT_HOME))
		fail("chdir: %s: %m", DEFAULT_HOME);

	/*
	 * We should set the environment accordingly.
	 */

	rc = pam_misc_setenv(pamh, "HOME", pw_dir, 0);
	if (rc != PAM_SUCCESS)
		fail("pam_misc_setenv: HOME");

	/*
	 * and now copy the environment for non-PAM use
	 */

	shell_env = pam_getenvlist(pamh);
	if (!shell_env)
		fail("pam_getenvlist: corrupt environment");

	rc = pam_end(pamh, PAM_SUCCESS);
	pamh = 0;
	if (rc != PAM_SUCCESS)
		fail("pam_end: failed to release authenticator");

	openlog("consolehelper", LOG_PID, LOG_AUTH);
	syslog(LOG_INFO, "executing \"%s\": (%s --> %s --> %s)",
	       conf->program, get_logname(), conf->user, conf->runas);
	closelog();

	cleanup_fds();

	/* assume user's identity */
	if (setuid(uid) < 0)
		fail("setuid(%d): %m", uid);

	signal(SIGCHLD, SIG_DFL);

	execve(conf->program, conf->argv, shell_env);
	fail("execve: %s: %m", conf->program);
}

static void
do_pam_init(const appconf * conf)
{
	int     rc;

	conv.appdata_ptr = (void *) conf->user;
	rc = pam_start(__progname, conf->user, &conv, &pamh);
	if (rc != PAM_SUCCESS)
		fail("auth: pam_start(%s,%s) failed with code %d",
		     __progname, conf->user, rc);

	rc = make_env(pamh, 0);
	if (rc != PAM_SUCCESS)
		fail("auth: problem establishing environment");

	/*
	 * set PAM_TTY
	 */
	{
		const char *term = ttyname(STDIN_FILENO);

		if (term)
		{
			rc = pam_set_item(pamh, PAM_TTY, (const void *) term);
			if (rc != PAM_SUCCESS)
				fail("auth: pam_set_item: PAM_TTY");
		}
	}

	/*
	 * set PAM_RUSER
	 */
	{
		const char *ruser = getlogin();

		if (ruser)
		{
			rc = pam_set_item(pamh, PAM_RUSER,
					  (const void *) ruser);
			if (rc != PAM_SUCCESS)
				fail("auth: pam_set_item: PAM_RUSER");
		}
	}

	/*
	 * set PAM_RHOST
	 */
	rc = pam_set_item(pamh, PAM_RHOST, (const void *) "localhost");
	if (rc != PAM_SUCCESS)
		fail("auth: pam_set_item: PAM_RHOST");
}

int
do_auth(const appconf * conf)
{
	int     rc;

	do_pam_init(conf);

	/* authenticate the user */
	rc = pam_authenticate(pamh, 0);
	if (rc != PAM_SUCCESS)
	{
		const char *msg = pam_strerror(pamh, rc);
		size_t  len = strlen(msg);

		write_client(msg, len, 1);
		fail("pam_authenticate failed: %s", msg);
	}

	rc = pam_acct_mgmt(pamh, 0);
	if (rc != PAM_SUCCESS)
	{
		const char *msg = pam_strerror(pamh, rc);
		size_t  len = strlen(msg);

		write_client(msg, len, 1);
		fail("pam_acct_mgmt failed: %s", msg);
	}

	if (conf->session)
	{
		pid_t   child;
		int     status;
		uid_t   uid;

		rc = pam_open_session(pamh, 0);
		if (rc != PAM_SUCCESS)
		{
			const char *msg = pam_strerror(pamh, rc);
			size_t  len = strlen(msg);

			write_client(msg, len, 1);
			fail("failed to open session: %s", msg);
		}

		/*
		 * Obtain all of the new credentials of the user
		 */
		rc = set_user_credentials(pamh, conf->runas, &uid);
		if (rc != PAM_SUCCESS)
		{
			const char *msg = pam_strerror(pamh, rc);
			size_t  len = strlen(msg);

			pam_close_session(pamh, rc);
			write_client(msg, len, 1);
			fail("failed to set credentials: %s",
			     pam_strerror(pamh, rc));
		}

		/*
		 * this is where we execute the program
		 */
		child = fork();
		if (child < 0)
		{
			pam_close_session(pamh, PAM_SYSTEM_ERR);
			fail("fork failed: %m");
		}

		if (!child)
			exec_program(uid, conf);

		close_client();

		/* wait for child to terminate */
		status = wait_for_child(child);

		rc = pam_setcred(pamh, PAM_DELETE_CRED);
		if (rc != PAM_SUCCESS)
			fail("failed to delete credentials: %s",
			     pam_strerror(pamh, rc));

		rc = pam_close_session(pamh, 0);
		if (rc != PAM_SUCCESS)
			fail("failed to close session: %s",
			     pam_strerror(pamh, rc));

		pam_end(pamh, rc);
		pamh = 0;

		return status;
	} else
	{
		uid_t   uid;

		/*
		 * Obtain all of the new credentials of the user
		 */
		rc = set_user_credentials(pamh, conf->runas, &uid);
		if (rc != PAM_SUCCESS)
		{
			fail("failed to set credentials: %s",
			     pam_strerror(pamh, rc));
		}

		/*
		 * just execute the program
		 */
		exec_program(uid, conf);
	}
}
