
/*
 * shvar.c
 *
 * Implementation of non-destructively reading/writing files containing
 * only shell variable declarations and full-line comments.
 *
 * Includes explicit inheritance mechanism intended for use with
 * Red Hat Linux ifcfg-* files.  There is no protection against
 * inheritance loops; they will generally cause stack overflows.
 * Furthermore, they are only intended for one level of inheritance;
 * the value setting algorithm assumes this.
 *
 * Copyright 1999,2000 Red Hat, Inc.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

#include "shvar.h"

/* remove escaped characters in place */
static void
unescape(char *s)
{
	int     len, i;

	len = strlen(s);
	if ((s[0] == '"' || s[0] == '\'') && s[0] == s[len - 1])
	{
		i = len - 2;
		memmove(s, s + 1, i);
		s[i + 1] = '\0';
		len = i;
	}
	for (i = 0; i < len; i++)
	{
		if (s[i] == '\\')
		{
			memmove(s + i, s + i + 1, len - (i + 1));
			len--;
		}
		s[len] = '\0';
	}
}

/* Open the file <name>, return shvarFile on success, NULL on failure */
shvarFile *
svNewFile(const char *name)
{
	int rb ,ro = 0;
	shvarList *lt;
	shvarFile *s = calloc(sizeof(*s), 1);

	if (!s)
		return NULL;

	s->fd = open(name, O_RDONLY);

	if (s->fd != -1)
	{
		struct stat buf;
		char   *p, *q, *e;

		if (fstat(s->fd, &buf) < 0)
			goto bail;
		s->arena = calloc(buf.st_size + 1, 1);

		do {
			rb = read(s->fd, s->arena + ro, buf.st_size - ro);
			if (rb < 0) {
				if (errno == EINTR)
					continue;
				goto bail;
			}
			ro += rb;
		} while (ro < buf.st_size);

		for (p = s->arena; (q = strchr(p, '\n')) != NULL; p = q + 1)
		{
			*q = '\0';

			e = strchr(p, '=');
			if (!e || *(e + 1) == '\0')
				continue;

			*e = '\0';
			unescape(e + 1);

			lt = calloc(sizeof(*lt), 1);
			if (!lt) {
				svCloseFile(s);
				return NULL;
			}

			lt->key = p;
			lt->value = e + 1;
			lt->next = s->lineList;

			s->lineList = lt;
		}

		return s;
	}

      bail:
	if (s->fd != -1)
		close(s->fd);
	free(s->arena);
	free(s);
	return NULL;
}

/* Get the value associated with the key.
 * The char* returned MUST be freed by the caller.
 */
const char   *
svGetValue(shvarFile * s, const char *key)
{
	shvarList *current;

	assert(s);
	assert(key);

	for (current = s->lineList; current;
	     current = current->next)
	{
		if (!strcmp(key, current->key))
		{
			return current->value;
		}
	}

	return NULL;
}

/* return 1 if <key> resolves to any truth value (e.g. "yes", "y", "true")
 * return 0 if <key> resolves to any non-truth value (e.g. "no", "n", "false")
 * return <default> otherwise
 */
int
svTrueValue(shvarFile * s, const char *key, int def)
{
	const char   *tmp;
	int     returnValue = def;

	tmp = svGetValue(s, key);
	if (!tmp)
		return returnValue;

	if ((!strcasecmp("yes", tmp)) ||
	    (!strcasecmp("true", tmp)) ||
	    (!strcasecmp("t", tmp)) || (!strcasecmp("y", tmp)))
		returnValue = 1;
	else if ((!strcasecmp("no", tmp)) ||
		 (!strcasecmp("false", tmp)) ||
		 (!strcasecmp("f", tmp)) || (!strcasecmp("n", tmp)))
		returnValue = 0;

	return returnValue;
}

/* Close the file descriptor (if open) and delete the shvarFile.
 * Returns -1 on error and 0 on success.
 */
int
svCloseFile(shvarFile * s)
{
	shvarList *current, *next;

	assert(s);

	if (s->fd != -1)
		close(s->fd);

	free(s->arena);
	for (current = s->lineList; current; current = next)
	{
		next = current->next;
		free(current);
	}
	free(s);
	return 0;
}
